package org.apache.ojb.odmg.collections;

/* Copyright 2002-2004 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.io.Serializable;
import java.util.AbstractMap;
import java.util.Iterator;

import org.apache.ojb.broker.PBKey;
import org.apache.ojb.broker.PersistenceBroker;
import org.apache.ojb.broker.PersistenceBrokerException;
import org.apache.ojb.broker.metadata.FieldDescriptor;
import org.apache.ojb.broker.metadata.fieldaccess.PersistentFieldFactory;
import org.apache.ojb.broker.util.collections.ManageableHashSet;
import org.apache.ojb.broker.util.logging.LoggerFactory;
import org.apache.ojb.odmg.PBCapsule;
import org.apache.ojb.odmg.TransactionImpl;
import org.apache.ojb.odmg.TxManagerFactory;
import org.odmg.DMap;
import org.odmg.Transaction;

/**
 * @author <a href="mailto:thma@apache.org">Thomas Mahler<a>
 * @version $Id: DMapImpl.java,v 1.21 2004/04/04 23:53:39 brianm Exp $
 */

public class DMapImpl extends AbstractMap implements DMap, Serializable
{
	private static final long serialVersionUID = 7048246616243056480L;
    private int id = -1;
    private ManageableHashSet entries;
    private int size = 0;
    private PBKey pbKey;

    /**
     * DMapImpl constructor comment.
     */
    public DMapImpl()
    {
        this.entries = new ManageableHashSet();
		if (!PersistentFieldFactory.usesAccessorsAndMutators())
	        id = generateNewId();
    }

    /**
     * DListImpl constructor comment.
     */
    public DMapImpl(PBKey key)
    {
        this.entries = new ManageableHashSet();
        this.pbKey = key;
		if (!PersistentFieldFactory.usesAccessorsAndMutators())
	        id = generateNewId();
    }

    /**
     * DMapImpl constructor comment.
     */
    public DMapImpl(int theId, int theSize)
    {
        id = theId;
        this.size = theSize;
    }

    public PBKey getPBKey()
    {
        return pbKey;
    }

    public void setPBKey(PBKey pbKey)
    {
        this.pbKey = pbKey;
    }

    /**
     * Returns a set view of the mappings contained in this map.  Each element
     * in the returned set is a <tt>Map.Entry</tt>.  The set is backed by the
     * map, so changes to the map are reflected in the set, and vice-versa.
     * If the map is modified while an iteration over the set is in progress,
     * the results of the iteration are undefined.  The set supports element
     * removal, which removes the corresponding mapping from the map, via the
     * <tt>Iterator.remove</tt>, <tt>Set.remove</tt>, <tt>removeAll</tt>,
     * <tt>retainAll</tt> and <tt>clear</tt> operations.  It does not support
     * the <tt>add</tt> or <tt>addAll</tt> operations.
     *
     * @return a set view of the mappings contained in this map.
     */
    public java.util.Set entrySet()
    {
        return entries;
    }

    /**
     * return a unique id
     */
    protected int generateNewId()
    {
        PBCapsule capsule = new PBCapsule(null, TxManagerFactory.instance().getTransaction());
        try
        {
            PersistenceBroker broker = capsule.getBroker();
            FieldDescriptor fld = broker.getClassDescriptor(this.getClass()).getAutoIncrementFields()[0];
            Integer val = (Integer) broker.serviceSequenceManager().getUniqueValue(fld);

            int result = val.intValue();
            return result;
        }
        catch (Exception e)
        {
            LoggerFactory.getDefaultLogger().error("DMapImpl: Generation of new id failed", e);
            throw new PersistenceBrokerException(e);
        }
        finally
        {
            capsule.destroy();
        }
    }


    /**
     * lazily retrieve the ID of the set, no need to precompute it.
     */
    public int getId()
    {
    	if (id == -1)
    	{
    		id = generateNewId();
    	}
        return id;
    }

    /**
     *
     */
    public Object put(Object key, Object value)
    {

        DMapEntry entry = new DMapEntry(this, key, value);
        boolean ok = entries.add(entry);
        if (ok)
        {
            TransactionImpl tx = TxManagerFactory.instance().getTransaction();
            if ((tx != null) && (tx.isOpen()))
            {
                tx.lock(this, Transaction.WRITE);
                tx.lock(entry, Transaction.WRITE);
                tx.lock(key, Transaction.READ);
                tx.lock(value, Transaction.READ);
            }
            return null;
        }
        else
        {
            return this.get(key);
        }
    }


    public Object remove(Object key)
    {
        Iterator i = entrySet().iterator();
        DMapEntry correctEntry = null;
        if (key == null)
        {
            while (correctEntry == null && i.hasNext())
            {
                DMapEntry e = (DMapEntry) i.next();
                if (e.getKey() == null)
                    correctEntry = e;
            }
        }
        else
        {
            while (correctEntry == null && i.hasNext())
            {
                DMapEntry e = (DMapEntry) i.next();
                if (key.equals(e.getKey()))
                    correctEntry = e;
            }
        }

        Object oldValue = null;
        if (correctEntry != null)
        {
            oldValue = correctEntry.getValue();
            i.remove();
            TransactionImpl tx = TxManagerFactory.instance().getTransaction();
            if ((tx != null) && (tx.isOpen()))
            {
                tx.markDelete(correctEntry);
            }
        }
        return oldValue;
    }


    /**
     * Gets the entries.
     * @return Returns a ManageableHashSet
     */
    public ManageableHashSet getEntries()
    {
        return entries;
    }

    /**
     * Sets the entries.
     * @param entries The entries to set
     */
    public void setEntries(ManageableHashSet entries)
    {
        this.entries = entries;
    }

    /**
     * Sets the id.
     * @param id The id to set
     */
    public void setId(int id)
    {
        this.id = id;
    }

    /**
     * Gets the size.
     * @return Returns a int
     */
    public int getSize()
    {
        return size;
    }

    /**
     * Sets the size.
     * @param size The size to set
     */
    public void setSize(int size)
    {
        this.size = size;
    }

}
